import tqdm
from logging import *
from termcolor import colored

# default level
getLogger().setLevel(INFO)

def merge(lhs, rhs):
    if type(lhs) != type(rhs):
        raise RuntimeError(f'cannot merge {type(lhs)} with {rhs}')
    if type(lhs) == dict:
        for k in rhs:
            if k in lhs:
                lhs[k] = merge(lhs[k], rhs[k])
            else:
                lhs[k] = rhs[k]
    elif type(lhs) == list:
        lhs.extend(rhs)
    return lhs


class bind_back:
    def __init__(self, fn, *args):
        self.fn = fn
        self.args = args

    def __call__(self, *args, **kwds):
        return self.fn(*args, *self.args, **kwds)


class ColoredFormatter(Formatter):

    COLORS = {
        'WARNING': bind_back(colored, 'yellow'),
        'INFO': bind_back(colored, 'green'),
        'DEBUG': bind_back(colored, 'cyan'),
        'CRITICAL': bind_back(colored, 'yellow'),
        'ERROR': bind_back(colored, 'red')
    }

    COLORS_ATTRS = {
        'WARNING': list(),
        'INFO': list(),
        'DEBUG': list(),
        'CRITICAL': ['blink'],
        'ERROR': ['blink']
    }

    COLOR_FORMAT = \
        f"[{colored('%(asctime)s.%(msecs)03d', 'grey')}]" \
        f"[%(levelname)s] {colored('%(name)s:', 'white', attrs=['bold'])} %(message)s "# \
        # f"({colored('%(filename)s:%(lineno)d', 'grey')})"

    FORMAT = \
        "[%(asctime)s.%(msecs)03d]" \
        "[%(levelname)s] %(name)s: %(message)s " # \
        # "(%(filename)s:%(lineno)d)"

    def __init__(self, use_color=True):
        super().__init__(self.COLOR_FORMAT if use_color else self.FORMAT, datefmt='%H:%M:%S')
        self.use_color = use_color

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            record.levelname = self.COLORS[levelname](
                levelname, attrs=['bold', *self.COLORS_ATTRS[levelname]])
            record.msg = self.COLORS[levelname](record.msg, attrs=[])
        return super().format(record)


_color_formatter = ColoredFormatter()
_no_color_formatter = ColoredFormatter(use_color=False)


class TqdmHandler(Handler):
    def emit(self, record):
        tqdm.tqdm.write(self.format(record))


def setup_logger(logger: Logger):
    if not logger.hasHandlers():
        console = TqdmHandler()
        console.setFormatter(_color_formatter)
        logger.addHandler(console)
    else:
        for handler in logger.handlers:
            handler.setFormatter(_color_formatter if isinstance(
                handler, StreamHandler) else _no_color_formatter)


class ColoredLogger(Logger):

    def __init__(self, name):
        super().__init__(name)
        self.propagate = False
        setup_logger(self)


setLoggerClass(ColoredLogger)


class Logging:
    def __init__(self, name: str = None) -> None:
        if name is None:
            name = self.__class__.__name__
        if name.startswith('root.'):
            name = name.removeprefix('root.')
        self._logger = getLogger(name)
        self.debug = self._logger.debug
        self.info = self._logger.info
        self.warning = self._logger.warning
        self.error = self._logger.error


def _get_makefile_logger():
    from dan.core.include import context
    if not hasattr(context.current, '_logger'):
        makefile_logger = getLogger(context.current.name)
        setup_logger(makefile_logger)
        setattr(context.current, '_logger', makefile_logger)
    else:
        makefile_logger: Logger = getattr(
            context.current, '_logger')
    return makefile_logger


def debug(*args, **kwds):
    return _get_makefile_logger().debug(*args, **kwds)


def info(*args, **kwds):
    return _get_makefile_logger().info(*args, **kwds)


def warning(*args, **kwds):
    return _get_makefile_logger().warning(*args, **kwds)


def error(*args, **kwds):
    return _get_makefile_logger().error(*args, **kwds)


def critical(*args, **kwds):
    return _get_makefile_logger().critical(*args, **kwds)


class lazy_fmt():    
    def __init__(self, fn):
        self.__fn=fn
    def __str__(self):
        return self.__fn()




import functools
import typing as t
from dan.core import asyncio
from dan.core.pm import re_match
from dan.core.settings import InstallMode, InstallSettings

from dan.core.version import VersionSpec
from dan.logging import Logging

def parse_package(name: str) -> tuple[str, str, str]:
    """Parse package name
    
    :returns: package, library, repository"""
    match re_match(name):
        # full specification <pkg>:<lib>@<repo>
        case r'(.+?):(.+?)@(.+)' as m:
            package = m[1]
            library = m[2]
            repository = m[3]
        # repo specification <lib>@<repo>
        case r'(.+?)@(.+)' as m:
            package = None
            library = m[1]
            repository = m[2]
        # package specification <pkg>:<lib>
        case r'(.+?):(.+)' as m:
            package = m[1]
            library = m[2]
            repository = None
        # no specification, automatic resolution in default repository
        case _:
            package = None
            library = name
            repository = None
    
    return package, library, repository


class RequiredPackage(Logging):
    def __init__(self, name: str, version_spec: VersionSpec = None):
        self.version_spec = version_spec
        super().__init__(name)
        self.target : 'Target' = None
        self.package, self.name, self.repository = parse_package(name)
        self.__skipped = list()

    def is_compatible(self, t: 'Target'):
        if self.version_spec is not None:
            version_ok = self.version_spec.is_compatible(t.version)
        else:
            version_ok = True
        return version_ok
    
    @property
    def found(self):
        return self.target is not None
    
    @property
    def modification_time(self):
        return self.target.modification_time if self.target else 0.0

    def __skipped_method_call(self, name, *args, **kwargs):
        if name not in self.__skipped:
            self.debug('call to %s skipped (unresolved)', name)
            self.__skipped.append(name)

    def __getattr__(self, name):
        if not self.found:
            return functools.partial(self.__skipped_method_call, name)
        else:
            return getattr(self.target, name)
    
    def __str__(self) -> str:
        if self.version_spec:
            return f'RequiredPackage[{self.name} {self.version_spec}]'
        return f'RequiredPackage[{self.name}]'
    

def parse_requirement(req: str) -> RequiredPackage:
    req = req.strip()
    name, spec = VersionSpec.parse(req)
    if spec:
        return RequiredPackage(name, spec)
    else:
        return RequiredPackage(req)

async def load_requirements(requirements: t.Iterable[RequiredPackage], makefile, logger = None, install = True):

    from dan.pkgconfig.package import find_package
    from dan.logging import _get_makefile_logger
    from dan.io import Package

    if logger is None:
        logger = _get_makefile_logger()

    deps_install_path = makefile.root.pkgs_path
    deps_settings = InstallSettings(deps_install_path)

    pkgs_search_paths = [deps_install_path]
    if makefile.requirements:
        pkgs_search_paths.append(makefile.requirements.pkgs_path)

    resolved: list[RequiredPackage] = list()
    unresolved: list[RequiredPackage] = list()

    async with asyncio.TaskGroup('requirement loading') as group:
        for req in requirements:
            if req.found:
                resolved.append(req)
                continue

            t = makefile.context.root.find(req.name)
            if t and not t.is_requirement and req.is_compatible(t):
                logger.debug('%s: already fullfilled by %s', req, t.fullname)
                req.target = t
                resolved.append(req)
                continue

            t = find_package(req.name, req.version_spec, search_paths=pkgs_search_paths, makefile=makefile)
            if t is not None and req.is_compatible(t):
                logger.debug('%s: using package %s', req, t.fullname)
                req.target = t
                resolved.append(req)
            elif install:
                if makefile.requirements:
                    # install requirement from dan-requires.py
                    t = makefile.requirements.find(req.name)
                    if not t:
                        raise RuntimeError(f'Unresolved requirement {req}, it should have been defined in {makefile.requirements.__file__}')
                    logger.debug('%s using requirements\' target %s', req, t.fullname)
                else:
                    with makefile.context:
                        t, is_new = Package.instance(req.name, req.version_spec, package=req.package, repository=req.repository, makefile=makefile.root)
                    if is_new:
                        logger.debug('%s: adding package %s', req, t.fullname)
                    else:
                        logger.debug('%s: package already beeing installed at version %s', req, t.version)
                group.create_task(t.install(deps_settings, InstallMode.dev))
                unresolved.append(req)



    if install:
        for req in unresolved:
            pkg = find_package(req.name, req.version_spec, search_paths=pkgs_search_paths, makefile=makefile)
            if pkg is None:
                raise RuntimeError(f'Unresolved requirement {req}')
            req.target = pkg
            resolved.append(req)

    return [req.target for req in resolved]




from dataclasses import dataclass, field
from enum import Enum
import functools
import itertools
import logging
import os
import fnmatch
import re

from dataclasses_json import dataclass_json
import sys
import tqdm
import typing as t
from collections.abc import Iterable

from dan.core import diagnostics as diag
from dan.core.cache import Cache
from dan.core.makefile import MakeFile
from dan.core.pathlib import Path
from dan.core.include import MakeFileError, include_makefile, Context
from dan.core import aiofiles, asyncio
from dan.core.settings import InstallMode, Settings
from dan.core.test import Test
from dan.core.utils import unique
from dan.cxx import init_toolchains
from dan.logging import Logging
from dan.core.target import Option, Target
from dan.cxx.targets import Executable
from dan.core.runners import max_jobs


def flatten(list_of_lists):
    if len(list_of_lists) == 0:
        return list_of_lists
    if isinstance(list_of_lists[0], Iterable):
        return flatten(list_of_lists[0]) + flatten(list_of_lists[1:])
    return list_of_lists[:1] + flatten(list_of_lists[1:])

@dataclass_json
@dataclass
class Config:
    source_path: Path = None
    build_path: Path = None
    toolchain: str = None
    settings: Settings = field(default_factory=lambda: Settings())

class ConfigCache(Cache[Config]):
    indent = 4

def _get_code_position(code, instruction_index):
    if instruction_index < 0 \
        or not hasattr(code, 'co_positions'): # Python <= 3.10 does not have co_position
        return (None, 0, 0, 0)
    positions_gen = code.co_positions()
    return next(itertools.islice(positions_gen, instruction_index // 2, None))

def _tb_entries(tb):
    entries = list()
    while tb is not None:
        entries.append(tb)
        tb = tb.tb_next
    return entries


def _walk_tb(tb, reverse=True):
    if isinstance(tb, Exception):
        tb = tb.__traceback__

    entries = _tb_entries(tb)
    
    if reverse:
        entries = reversed(entries)
    
    for tb in entries:
        positions = _get_code_position(tb.tb_frame.f_code, tb.tb_lasti)
        # Yield tb_lineno when co_positions does not have a line number to
        # maintain behavior with walk_tb.
        if positions[0] is None:
            yield tb.tb_frame.f_code.co_filename, (tb.tb_lineno, ) + positions[1:]
        else:
            yield tb.tb_frame.f_code.co_filename, positions


def gen_python_diags(err: Exception):
    diagnostics = diag.DiagnosticCollection()
    if not diag.enabled:
        return diagnostics
    match err:
        case MakeFileError():
            cause = err.__cause__
            related = None
            if isinstance(cause, MakeFileError):
                related = list()
                cause_diags = gen_python_diags(cause)
                for filename, diagns in cause_diags.items():
                    for d in diagns:
                        related.append(diag.RelatedInformation(diag.Location(diag.Uri(filename), d.range), d.message))
                diagnostics.update(cause_diags)
                err_filename = str(err.path)
                for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(cause):
                    if filename == err_filename:
                        break
            else:
                for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(cause):
                    break
            diagnostics[filename] = diag.Diagnostic(
                message=str(cause),
                range=diag.Range(start=diag.Position(lineno - 1, colno), end=diag.Position(end_lineno - 1, end_colno)),
                code=cause.__class__.__name__,
                source='dan.make',
                related_information=related
            )
        case _:
            for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(err):
                break
            diagnostics[filename] = diag.Diagnostic(
                message=str(err),
                range=diag.Range(start=diag.Position(lineno - 1, colno), end=diag.Position(end_lineno - 1, end_colno)),
                code=err.__class__.__name__,
                source='dan.make'
            )

    return diagnostics

class Make(Logging):
    _config_name = 'dan.config.json'
    _cache_name = 'dan.cache'

    def __init__(self, build_path: str, targets: list[str] = None, verbose: bool = False, quiet: bool = False, for_install: bool = False, jobs: int = None, no_progress=False, diags=False):

        jobs = jobs or os.cpu_count()
        max_jobs(jobs)

        if quiet:
            assert not verbose, "'quiet' cannot be combined with 'verbose'"
            log_level = logging.ERROR
        elif verbose:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logging.getLogger().setLevel(log_level)

        super().__init__('make')

        if diags:
            diag.enabled = True

        self.no_progress = no_progress
        self.for_install = for_install
        
        self.build_path = Path(build_path)
        self.config_path = build_path / self._config_name
        self.cache_path = build_path / self._cache_name

        self.required_targets = targets
        sys.pycache_prefix = str(build_path / '__pycache__')
        self._config = ConfigCache.instance(self.config_path)
        self.cache = Cache.instance(self.cache_path, binary=True)
        
        self.debug(f'jobs: {jobs}')

        self._diagnostics = diag.DiagnosticCollection()

        self.context = Context()
        
    @property
    def config(self) -> Config:
        return self._config.data
    
    @property
    def settings(self) -> Settings:
        return self._config.data.settings
    
    @property
    def source_path(self):
        return Path(self.config.source_path)
        
    @property
    def toolchain(self):
        return self.context.get('cxx_target_toolchain')
    
    @property
    def root(self) -> MakeFile:
        return self.context.root

    async def configure(self, source_path: str, toolchain: str = None):
        self.config.source_path = str(source_path)
        self.config.build_path = str(self.build_path)
        self.info(f'source path: {self.config.source_path}')
        self.info(f'build path: {self.config.build_path}')
        if toolchain:
            self.config.toolchain = toolchain
        if not self.config.toolchain:
            self.warning('no toolchain configured')
        await self._config.save()

    @asyncio.cached
    async def initialize(self):
        assert self.config_path.exists(), 'configure first'

        self.debug(f'source path: {self.source_path}')
        self.debug(f'build path: {self.build_path}')

        toolchain = self.config.toolchain
        build_type = self.settings.build_type

        self.info(
            f'using \'{toolchain}\' toolchain in \'{build_type.name}\' mode')

        with self.context:
            init_toolchains(toolchain, self.settings)
            try:
                include_makefile(self.source_path, self.build_path)
            except MakeFileError as err:
                self._diagnostics.update(gen_python_diags(err))
                raise

        target_toolchain = self.context.get('cxx_target_toolchain')
        target_toolchain.build_type = build_type
        if self.for_install:
            library_dest = Path(self.settings.install.destination) / \
                self.settings.install.libraries_prefix
            target_toolchain.rpath = str(library_dest.absolute())

        self.debug(f'targets: {[t.name for t in self.targets]}')

    def __matches(self, target: Target | Test):
        for required in self.required_targets:
            if fnmatch.fnmatch(target.fullname, f'*{required}*'):
                return True
        return False

    @functools.cached_property
    def targets(self) -> list[Target]:
        items = list()
        if self.required_targets and len(self.required_targets) > 0:
            for target in self.root.all_targets:
                if self.__matches(target):
                    items.append(target)
        else:
            items = self.root.all_default
        return items

    @functools.cached_property
    def tests(self) -> list[Test]:
        items = list()
        if self.required_targets and len(self.required_targets) > 0:
            for required in self.required_targets:
                pos = required.find(':')
                if pos > 0:
                    test_name = required[:pos]
                    test_case = required[pos+1:]
                else:
                    test_name = required
                    test_case = None

                for test in self.root.all_tests:
                    
                    if fnmatch.fnmatch(test.fullname, f'*{test_name}*'):
                        if len(test) > 1 and test_case is not None:
                            cases = list()
                            for case in test.cases:
                                if fnmatch.fnmatch(case.name, test_case):
                                    cases.append(case)
                            if len(cases) == 0:
                                self.warning('couldn\'t find any test case in %s matching \'%s\'', test.name, test_case)
                            test.cases = cases
                        
                        items.append(test)
        else:
            for test in self.root.all_tests:
                items.append(test)
        return items

    @property
    def all_options(self) -> list[Option]:
        opts = []
        for target in self.targets:
            for o in target.options:
                opts.append(o)
        for makefile in self.context.all_makefiles:
            for o in makefile.options:
                opts.append(o)
        return opts
    
    @property
    def diagnostics(self):
        result: diag.DiagnosticCollection = diag.DiagnosticCollection()
        if self.root:
            for target in self.root.all_targets:
                result.update(target.diagnostics)
        result.update(self._diagnostics)
        return result


    async def target_of(self, source: Path):
        from dan.cxx.targets import CXXObjectsTarget

        source = Path(source)
        if source.suffix[1].lower() == 'h':
            def check(t: CXXObjectsTarget):
                for p in [*t.includes.private_raw, *t.includes.public_raw]:
                    p: Path = p
                    if p not in source.parents:
                        continue
                    return True
                return False
        else:
            def check(t: CXXObjectsTarget):
                t._init_sources()
                if source.name in [Path(s).name for s in target.sources]:
                    return True
                return False
        for target in [target for target in self.root.all_targets if isinstance(target, CXXObjectsTarget)]:            
            if target.source_path not in source.parents:
                continue
            if check(target):
                return target


    @classmethod
    def _parse_str_value(cls, name, value: str, orig: type, tp: type = None):
        if issubclass(orig, Enum):
            names = [n.lower()
                        for n in orig._member_names_]
            value = value.lower()
            if value in names:
                return orig(names.index(value))
            else:
                raise RuntimeError(f'{name} should be one of {names}')
        elif issubclass(orig, (set, list)):
            assert tp is not None
            result = list()
            for sub in value.split(';'):
                result.append(cls._parse_str_value(name, sub, tp))
            return orig(result)
        elif orig == bool:
            return value.lower() in ('true', 'yes', 'on', '1')
        else:
            if tp is not None:
                raise TypeError(f'unhandled type {orig}[{tp}]')
            return orig(value)

    
    @classmethod
    def _apply_inputs(self, inputs: list[str], get_item: t.Callable[[str], tuple[t.Any, t.Any, t.Any]], info: t.Callable[[t.Any], t.Any]):
        for input in inputs:
            m = re.match(r'(.+?)([+-])?="?(.+)"?', input)
            if m:
                name = m[1]
                op = m[2]
                value = m[3]
                input, out_value, orig = get_item(name)
                sname = name.split('.')[-1]
                if orig is None:
                    orig = type(input)
                if hasattr(orig, '__annotations__') and sname in orig.__annotations__:
                    tp = orig.__annotations__[sname]
                    orig = t.get_origin(tp)
                    if orig is None:
                        orig = tp
                        tp = None
                    else:
                        args = t.get_args(tp)
                        if args:
                            tp = args[0]
                else:
                    tp = None
                in_value = self._parse_str_value(name, value, orig, tp)
                match (out_value, op, in_value):
                    case (list()|set(), '-', list()|set()) if len(in_value) == 1 and list(in_value)[0] == '*':
                        out_value.clear()
                    case (set(), '+', set()):
                        out_value.update(in_value)
                    case (set(), '-', set()):
                        out_value = out_value - in_value
                    case (list(), '+', set()|list()):
                        out_value.extend(in_value)
                    case (list(), '-', set()|list()):
                        for v in in_value:
                            out_value.remove(v)
                    case (_, '+' | '-', _):
                        raise TypeError(f'unhandled "{op}=" operator on type {type(out_value)} ({name})')
                    case _:
                        out_value = in_value
                if isinstance(input, dict):
                    input[sname] = out_value
                else:
                    setattr(input, sname, out_value)
                info(name, out_value)
            else:
                raise RuntimeError(f'cannot process given input: {input}')

    async def apply_options(self, *options):
        await self.initialize()
        all_opts = self.all_options
        def get_option(name):
            for opt in all_opts:
                if opt.fullname == name:
                    return opt.cache, opt.value, opt.type
        self._apply_inputs(options, get_option, lambda k, v: self.info(f'option: {k} = {v}'))

    async def apply_settings(self, *settings):
        # dont init for settings
        def get_setting(name):
            parts = name.split('.')
            setting = self.settings
            for part in parts[:-1]:
                if not hasattr(setting, part):
                    raise RuntimeError(f'no such setting: {name}')
                setting = getattr(setting, part)
            if not hasattr(setting, parts[-1]):
                raise RuntimeError(f'no such setting: {name}')
            value = getattr(setting, parts[-1])
            return setting, value, type(setting)
        self._apply_inputs(settings, get_setting, lambda k, v: self.info(f'setting: {k} = {v}'))


    @staticmethod
    def toolchains():
        from dan.cxx.detect import get_toolchains
        return get_toolchains()

    class progress:

        def __init__(self, desc, targets, task_builder, disable) -> None:
            self.desc = desc
            self.targets = list(targets) # NOTE: need to take a copy of the list
            self.builder = task_builder
            import shutil
            term_cols = shutil.get_terminal_size().columns
            self.max_desc_width = int(term_cols * 0.25)
            self.pbar = tqdm.tqdm(total=len(targets),
                                  desc='building', disable=disable)
            self.pbar.unit = ' targets'

        def __enter__(self):
            def update(n=1):
                desc = self.desc + ' ' + \
                    ', '.join([t.name for t in self.targets])
                if len(desc) > self.max_desc_width:
                    desc = desc[:self.max_desc_width] + ' ...'
                self.pbar.set_description_str(desc)
                self.pbar.update(n)
            update(0)

            def on_done(t: Target, *args, **kwargs):
                self.targets.remove(t)
                update()

            tasks = list()

            for t in self.targets:
                tsk = asyncio.create_task(self.builder(t))
                tsk.add_done_callback(functools.partial(on_done, t))
                tasks.append(tsk)

            return tasks

        def __exit__(self, *args):
            self.pbar.set_description_str(self.desc + ' done')
            self.pbar.refresh()
            return
        
    async def _build_target(self, t: Target):
        try:
            await t.build()
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def build(self, targets: list[Target] = None):
        await self.initialize()

        if targets is None:
            targets = self.targets

        with self.context, \
             self.progress('building', targets, self._build_target, self.no_progress) as tasks:
            results = await asyncio.gather(*tasks, return_exceptions=True)
            errors = list()
            for result in results:
                if isinstance(result, Exception):
                    self._logger.error(str(result))
                    errors.append(result)
            err_count = len(errors)
            if err_count == 1:
                raise errors[0]
            elif err_count > 1:
                raise asyncio.ExceptionGroup('Multiple errors occured while building the project', errors=errors)

    async def _install_target(self, t: Target, mode: InstallMode):
        try:
            return await t.install(self.settings.install, mode)
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def install(self, mode: InstallMode = InstallMode.user):
        await self.initialize()

        self.for_install = True
        await self.initialize()
        targets = []
        for t in self.targets:
            if t.installed:
                targets.append(t)
            else:
                self.debug('skipping %s (not marked as installed)', t.fullname)

        await self.build(targets)

        with self.context, \
             self.progress('installing', targets, functools.partial(self._install_target, mode=mode), self.no_progress) as tasks:
            installed_files = await asyncio.gather(*tasks)
            installed_files = unique(flatten(installed_files))
            manifest_path = self.settings.install.data_destination / \
                'dan' / f'{self.root.name}-manifest.txt'
            manifest_path.parent.mkdir(parents=True, exist_ok=True)

            async with aiofiles.open(manifest_path, 'w') as f:
                await f.writelines([os.path.relpath(p, manifest_path.parent) + '\n' for p in installed_files])

    @property
    def executable_targets(self) -> list[Executable]:
        return [exe for exe in self.targets if isinstance(exe, Executable)]

    async def scan_toolchains(self, script: Path = None):
        from dan.cxx.detect import create_toolchains, load_env_toolchain
        if script:
            load_env_toolchain(script)
        else:
            create_toolchains()

    async def run(self):
        await self.initialize()
        with self.context:
            results = await asyncio.gather(*[t.execute(log=True) for t in self.executable_targets])
            for result in results:
                if result[2] != 0:
                    return result[2]
        return 0

    async def _test_target(self, t: Test):
        try:
            return await t.run_test()
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def test(self):
        await self.initialize()
        tests = self.tests
        if len(tests) == 0:
            self.error('No test selected')
            return 255

        with self.context:
            with self.progress('testing', tests, self._test_target, self.no_progress) as tasks:
                results = await asyncio.gather(*tasks)
                if all(results):
                    self.info('Success !')
                    return 0
                else:
                    self.error('Failed !')
                    return 255

    async def clean(self):
        await self.initialize()
        with self.context:
            await asyncio.gather(*[t.clean() for t in self.targets])
            # from dan.cxx import target_toolchain
            # target_toolchain.compile_commands.clear()
